@@ -1,5 +1,4 @@
 import numpy as np
-import torch
 from ase.calculators.calculator import Calculator, all_changes
 
 
@@ -15,7 +14,7 @@ class GPRCalculator(Calculator):
         self.results = {}
 
     def calculate(self, atoms=None,
-                  properties=['energy', 'forces', 'uncertainty'],
+                  properties=['energy', 'forces', 'unc_energy', 'unc_forces'],
                   system_changes=all_changes):
         '''
         Calculate the energy and forces for a given Atoms structure.
@@ -25,10 +24,12 @@ class GPRCalculator(Calculator):
 
         Calculator.calculate(self, atoms, properties, system_changes)
 
-        energy_gpr, force_gpr, uncertainty_gpr = self.calculator.eval_per_data(eval_image=atoms, get_variance=True)
+        energy_gpr, force_gpr, unc_e_gpr, unc_f_gpr = self.calculator.eval_per_data(eval_image=atoms, get_variance=True)
+
         energy_gpr = energy_gpr.cpu().detach().numpy()
         force_gpr = force_gpr.cpu().detach().numpy()
-        uncertainty_gpr = uncertainty_gpr.cpu().detach().numpy()
+        unc_e_gpr = unc_e_gpr.cpu().detach().numpy()
+        unc_f_gpr = unc_f_gpr.cpu().detach().numpy()
 
         if self.train_data.standardization:
             mean_energy = np.mean(self.train_data.energy)
@@ -42,4 +43,5 @@ class GPRCalculator(Calculator):
 
         self.results['energy'] = energy_gpr
         self.results['forces'] = force_gpr
-        self.results['uncertainty'] = uncertainty_gpr
+        self.results['unc_energy'] = unc_e_gpr
+        self.results['unc_forces'] = unc_f_gpr
