@@ -9,72 +9,165 @@ from aenet_gpr.src.pytorch_kernel import FPKernel
 from aenet_gpr.util.prepare_data import get_N_batch, get_batch_indexes_N_batch
 
 
-# def _central_diff_task(atoms, i, j, delta, model, num_layers):
-#     """
-#     Desciptor gradient for each atom i, and cartexian coordinate j(x/y/z)
-#     """
-#     atoms_f = deepcopy(atoms)
-#     atoms_f.positions[i, j] += delta
-#     desc_f = model.get_descriptors(atoms_f, num_layers=num_layers)
-#
-#     atoms_b = deepcopy(atoms)
-#     atoms_b.positions[i, j] -= delta
-#     desc_b = model.get_descriptors(atoms_b, num_layers=num_layers)
-#
-#     grad_ij = (desc_f - desc_b) / (2 * delta)
-#     return i, j, grad_ij
-#
-#
-# def numerical_descriptor_gradient_parallel(atoms, model, delta=1e-4, num_layers=-1, n_jobs=1, dtype=torch.float32):
-#     """
-#     Args:
-#         atoms (ase.Atoms):
-#         model: MACE calculator
-#         delta (float): displacement (Ang)
-#         num_layers (int): MACE interaction layers number
-#         n_jobs (int): pararell job number
-#         dtype (torch.dtype):
-#
-#     Returns:
-#         desc (torch.Tensor): (n_atoms, descriptor_dim)
-#         grad (torch.Tensor): (n_atoms, n_atoms, 3, descriptor_dim)
-#     """
-#
-#     desc = model.get_descriptors(atoms, num_layers=num_layers)
-#     n_atoms, D = desc.shape
-#
-#     # Job list
-#     jobs = [(atoms, i, j, delta, model, num_layers) for i in range(n_atoms) for j in range(3)]
-#
-#     # Pararell execution
-#     results = Parallel(n_jobs=n_jobs, prefer="threads")(delayed(_central_diff_task)(*job) for job in jobs)
-#     results.sort(key=lambda x: (x[0], x[1]))
-#
-#     grad = torch.empty((n_atoms, n_atoms, 3, D), dtype=dtype)
-#     for i, j, grad_ij in results:
-#         grad[:, i, j, :] = torch.tensor(grad_ij, dtype=dtype)
-#
-#     return torch.tensor(desc, dtype=dtype), grad
+def chebyshev_descriptor_gradient(atoms, model, atoms_mask, delta=1e-4, n_jobs=-1, dtype='float32'):
+    """
+    Args:
+        atoms (ase.Atoms):
+        model: Chebyshev calculator
+        atoms_mask: Atom index to keep
+        delta (float): displacement (Ang)
+        n_jobs (int): number of processes
+        dtype (dtype):
+
+    Returns:
+        desc (torch.Tensor): (n_reduced_atoms, descriptor_dim)
+        grad (torch.Tensor): (n_reduced_atoms, n_reduced_atoms, 3, descriptor_dim)
+    """
+    # Get original descriptor
+    atoms_mask = atoms_mask.cpu().numpy()
+    if np.any(atoms.pbc):
+        desc = model.featurize_structure(atoms.positions,
+                                         atoms.get_chemical_symbols(),
+                                         cell=np.array([atoms.cell[i] for i in range(3)]),
+                                         pbc=atoms.pbc)
+        n_atoms, D = desc.shape
+
+        # Pre-allocate gradient array
+        grad = np.empty((n_atoms, n_atoms, 3, D), dtype=dtype)
+
+        # Get original positions once
+        original_positions = atoms.get_positions()
+
+        # Pre-create all perturbations
+        perturbations = []
+        for i in atoms_mask:
+            for j in range(3):  # x, y, z
+                # Forward perturbation
+                pos_f = original_positions.copy()
+                pos_f[i, j] += delta
+                perturbations.append(('f', i, j, pos_f))
+
+                # Backward perturbation
+                pos_b = original_positions.copy()
+                pos_b[i, j] -= delta
+                perturbations.append(('b', i, j, pos_b))
+
+        if n_jobs == 1:
+            # Compute all descriptors
+            all_descriptors = []
+            for direction, i, j, positions in perturbations:
+                atoms_temp = deepcopy(atoms)
+                atoms_temp.set_positions(positions)
+                desc_temp = model.featurize_structure(atoms.positions,
+                                                      atoms.get_chemical_symbols(),
+                                                      cell=np.array([atoms.cell[i] for i in range(3)]),
+                                                      pbc=atoms.pbc)
+                all_descriptors.append((direction, i, j, desc_temp))
+        else:
+            # --- Parallelize descriptor computation ---
+            def compute_descriptor(direction, i, j, positions):
+                atoms_temp = deepcopy(atoms)
+                atoms_temp.set_positions(positions)
+                desc_temp = model.featurize_structure(atoms.positions,
+                                                      atoms.get_chemical_symbols(),
+                                                      cell=np.array([atoms.cell[i] for i in range(3)]),
+                                                      pbc=atoms.pbc)
+                return direction, i, j, desc_temp
+
+            all_descriptors = Parallel(n_jobs=n_jobs, prefer="threads")(
+                delayed(compute_descriptor)(direction, i, j, positions)
+                for (direction, i, j, positions) in perturbations
+            )
+
+    else:
+        desc = model.featurize_structure(atoms.positions,
+                                         atoms.get_chemical_symbols())
+        n_atoms, D = desc.shape
+
+        # Pre-allocate gradient array
+        grad = np.empty((n_atoms, n_atoms, 3, D), dtype=dtype)
+
+        # Get original positions once
+        original_positions = atoms.get_positions()
+
+        # Pre-create all perturbations
+        perturbations = []
+        for i in atoms_mask:
+            for j in range(3):  # x, y, z
+                # Forward perturbation
+                pos_f = original_positions.copy()
+                pos_f[i, j] += delta
+                perturbations.append(('f', i, j, pos_f))
+
+                # Backward perturbation
+                pos_b = original_positions.copy()
+                pos_b[i, j] -= delta
+                perturbations.append(('b', i, j, pos_b))
+
+        if n_jobs == 1:
+            # Compute all descriptors
+            all_descriptors = []
+            for direction, i, j, positions in perturbations:
+                atoms_temp = deepcopy(atoms)
+                atoms_temp.set_positions(positions)
+                desc_temp = model.featurize_structure(atoms.positions,
+                                                      atoms.get_chemical_symbols())
+                all_descriptors.append((direction, i, j, desc_temp))
+        else:
+            # --- Parallelize descriptor computation ---
+            def compute_descriptor(direction, i, j, positions):
+                atoms_temp = deepcopy(atoms)
+                atoms_temp.set_positions(positions)
+                desc_temp = model.featurize_structure(atoms.positions,
+                                                      atoms.get_chemical_symbols())
+                return direction, i, j, desc_temp
+
+            all_descriptors = Parallel(n_jobs=n_jobs, prefer="threads")(
+                delayed(compute_descriptor)(direction, i, j, positions)
+                for (direction, i, j, positions) in perturbations
+            )
+
+    # Reorganize and compute gradients
+    desc_dict = {}
+    for direction, i, j, desc_temp in all_descriptors:
+        key = (i, j)
+        if key not in desc_dict:
+            desc_dict[key] = {}
+        desc_dict[key][direction] = desc_temp[:, :]
+
+    # Compute central differences
+    for (i, j), descs in desc_dict.items():
+        grad[:, i, j, :] = (descs['f'] - descs['b']) / (2 * delta)
+
+    # n_atoms -> n_reduced_atoms
+    desc = desc[atoms_mask, :]
+    grad = grad[atoms_mask, :, :, :]
+    grad = grad[:, atoms_mask, :, :]
+
+    return desc, grad
 
 
-def numerical_descriptor_gradient(atoms, model, delta=1e-4, num_layers=-1, n_jobs=-1, dtype='float32'):
+def mace_descriptor_gradient(atoms, model, atoms_mask, delta=1e-4, invariants=True, num_layers=-1, n_jobs=-1,
+                             dtype='float32'):
     """
     Optimized batch version - pre-generates all perturbed positions and processes them efficiently
 
     Args:
         atoms (ase.Atoms):
         model: MACE calculator
+        atoms_mask: Atom index to keep
         delta (float): displacement (Ang)
         num_layers (int): MACE interaction layers number
         n_jobs (int): number of processes
         dtype (dtype):
 
     Returns:
-        desc (torch.Tensor): (n_atoms, descriptor_dim)
-        grad (torch.Tensor): (n_atoms, n_atoms, 3, descriptor_dim)
+        desc (torch.Tensor): (n_reduced_atoms, descriptor_dim)
+        grad (torch.Tensor): (n_reduced_atoms, n_reduced_atoms, 3, descriptor_dim)
     """
     # Get original descriptor
-    desc = model.get_descriptors(atoms, num_layers=num_layers)
+    atoms_mask = atoms_mask.cpu().numpy()
+    desc = model.get_descriptors(atoms, invariants_only=invariants, num_layers=num_layers)
     n_atoms, D = desc.shape
 
     # Pre-allocate gradient array
@@ -85,7 +178,7 @@ def numerical_descriptor_gradient(atoms, model, delta=1e-4, num_layers=-1, n_job
 
     # Pre-create all perturbations
     perturbations = []
-    for i in range(n_atoms):
+    for i in atoms_mask:
         for j in range(3):  # x, y, z
             # Forward perturbation
             pos_f = original_positions.copy()
@@ -103,14 +196,14 @@ def numerical_descriptor_gradient(atoms, model, delta=1e-4, num_layers=-1, n_job
         for direction, i, j, positions in perturbations:
             atoms_temp = deepcopy(atoms)
             atoms_temp.set_positions(positions)
-            desc_temp = model.get_descriptors(atoms_temp, num_layers=num_layers)
+            desc_temp = model.get_descriptors(atoms_temp, invariants_only=invariants, num_layers=num_layers)
             all_descriptors.append((direction, i, j, desc_temp))
     else:
         # --- Parallelize descriptor computation ---
         def compute_descriptor(direction, i, j, positions):
             atoms_temp = deepcopy(atoms)
             atoms_temp.set_positions(positions)
-            desc_temp = model.get_descriptors(atoms_temp, num_layers=num_layers)
+            desc_temp = model.get_descriptors(atoms_temp, invariants_only=invariants, num_layers=num_layers)
             return direction, i, j, desc_temp
 
         all_descriptors = Parallel(n_jobs=n_jobs, prefer="threads")(
@@ -124,70 +217,36 @@ def numerical_descriptor_gradient(atoms, model, delta=1e-4, num_layers=-1, n_job
         key = (i, j)
         if key not in desc_dict:
             desc_dict[key] = {}
-        desc_dict[key][direction] = desc_temp
+        desc_dict[key][direction] = desc_temp[:, :]
 
     # Compute central differences
     for (i, j), descs in desc_dict.items():
         grad[:, i, j, :] = (descs['f'] - descs['b']) / (2 * delta)
 
+    # n_atoms -> n_reduced_atoms
+    desc = desc[atoms_mask, :]
+    grad = grad[atoms_mask, :, :, :]
+    grad = grad[:, atoms_mask, :, :]
+
     return desc, grad
 
 
-# def numerical_descriptor_gradient(atoms, model, delta=1e-4, num_layers=-1, dtype='float32'):
-#     """
-#     Args:
-#         atoms (ase.Atoms):
-#         model: MACE calculator
-#         delta (float): displacement (Ang)
-#         num_layers (int): MACE interaction layers number
-#         dtype (torch.dtype):
-#
-#     Returns:
-#         desc (torch.Tensor): (n_atoms, descriptor_dim)
-#         grad (torch.Tensor): (n_atoms, n_atoms, 3, descriptor_dim)
-#     """
-#
-#     desc = model.get_descriptors(atoms, num_layers=num_layers)
-#     n_atoms, D = desc.shape
-#
-#     grad = np.empty((n_atoms, n_atoms, 3, D), dtype=dtype)
-#     for i in range(n_atoms):
-#         for j in range(3):  # x, y, z
-#             # forward step
-#             atoms_f = deepcopy(atoms)
-#             atoms_f.positions[i, j] += delta
-#             desc_f = model.get_descriptors(atoms_f, num_layers=num_layers)
-#
-#             # backward step
-#             atoms_b = deepcopy(atoms)
-#             atoms_b.positions[i, j] -= delta
-#             desc_b = model.get_descriptors(atoms_b, num_layers=num_layers)
-#
-#             # central difference
-#             grad[:, i, j, :] = (desc_f - desc_b) / (2 * delta)
-#
-#     return desc, grad
-
-
-def apply_force_mask(F, atoms_mask):
+def apply_force_mask(F, atoms_xyz_mask):
     """
     Args:
         F: (Ntest, 3*Natoms) force tensor
-        atoms_mask: tensor([...]) flattened xyz indices to keep
+        atoms_xyz_mask: tensor([...]) flattened xyz indices to keep
 
     Returns:
         F_masked: same shape as F, masked with zeros outside atoms_mask
     """
-    if atoms_mask is None:
-        # keep all forces
-        return F.clone()
-    else:
-        mask = torch.zeros(F.shape[1], dtype=torch.bool, device=F.device)
-        mask[atoms_mask] = True
 
-        F_masked = torch.zeros_like(F)
-        F_masked[:, mask] = F[:, mask]
-        return F_masked
+    mask = torch.zeros(F.shape[1], dtype=torch.bool, device=F.device)
+    mask[atoms_xyz_mask] = True
+
+    F_masked = torch.zeros_like(F)
+    F_masked[:, mask] = F[:, mask]
+    return F_masked
 
 
 class GaussianProcess(object):
@@ -206,7 +265,7 @@ class GaussianProcess(object):
                  sparse=None, sparse_derivative=None, autograd=False,
                  train_batch_size=25, eval_batch_size=25,
                  data_type='float64', device='cpu',
-                 soap_param=None, mace_param=None, descriptor='cartesian coordinates',
+                 soap_param=None, mace_param=None, cheb_param=None, descriptor='cartesian coordinates',
                  atoms_mask=None):
 
         if data_type == 'float32':
@@ -220,6 +279,7 @@ class GaussianProcess(object):
 
         self.soap_param = soap_param
         self.mace_param = mace_param
+        self.cheb_param = cheb_param
         self.descriptor = descriptor
         self.kerneltype = kerneltype
 
@@ -234,13 +294,16 @@ class GaussianProcess(object):
         self.Ntrain = len(self.images)
         self.species = self.images[0].get_chemical_symbols()
         self.pbc = np.all(self.images[0].get_pbc())
-
         self.Natom = len(self.species)
-        self.atoms_mask = atoms_mask
 
         if self.descriptor == 'soap':
             try:
                 from dscribe.descriptors import SOAP
+                print("You are using SOAP descriptor through DScribe:")
+                print("[1] A. P. Bartók, R. Kondor and G. Csányi, Phys. Rev. B 87 (2013) 184115.")
+                print("[2] L. Himanen, A. S Foster et al., Comput. Phys. Commun. 247 (2020) 106949. \n")
+                print("SOAP parameter:")
+                print(self.soap_param)
             except ImportError:
                 raise ImportError(
                     "The 'dscribe' package is required for using SOAP descriptors.\n"
@@ -261,6 +324,12 @@ class GaussianProcess(object):
             if self.mace_param.get('system') == "materials":
                 try:
                     from mace.calculators import mace_mp
+                    print("You are using pre-trained MACE descriptor:")
+                    print(
+                        "[1] I. Batatia, D. P Kovacs, G. Simm, C. Ortner, and G. Csányi, Adv. Neural Inf. Process. Syst. 35 (2022) 11423.")
+                    print("[2] I. Batatia, G. Csányi et al., arXiv:2401.00096 (2023). \n")
+                    print("MACE parameter:")
+                    print(self.mace_param)
                 except ImportError:
                     raise ImportError(
                         "The 'joblib' and 'mace' packages are required for using pre-trained MACE descriptors.\n"
@@ -279,6 +348,31 @@ class GaussianProcess(object):
                 self.mace = mace_mp(model=self.mace_param.get('model'),
                                     device=mace_device)
 
+            elif self.descriptor == 'chebyshev':
+                try:
+                    from aenet.torch_featurize import ChebyshevDescriptor
+                    print("You are using Chebyshev descriptor:")
+                    print("N. Artrith, A. Urban, and G. Ceder, Phys. Rev. B 96 (2017) 014112. \n")
+                    print("Chebshev parameter:")
+                    print(self.cheb_param)
+                except ImportError:
+                    raise ImportError(
+                        "The 'aenet-python' package is required for using Chebyshev descriptors.\n"
+                        "Please install it by running:\n\n"
+                        "    git clone https://github.com/atomisticnet/aenet-python.git\n"
+                        "    cd ./aenet-python/\n"
+                        "    pip install . --user\n"
+                    )
+
+                self.cheb = ChebyshevDescriptor(species=set(self.species),
+                                                rad_order=self.cheb_param.get("rad_order"),  # Radial polynomial order
+                                                rad_cutoff=self.cheb_param.get("rad_cutoff"),
+                                                # Radial cutoff (Angstroms)
+                                                ang_order=self.cheb_param.get("ang_order"),  # Angular polynomial order
+                                                ang_cutoff=self.cheb_param.get("ang_cutoff")
+                                                # Angular cutoff (Angstroms)
+                                                )
+
             else:
                 try:
                     from mace.calculators import mace_off
@@ -293,14 +387,31 @@ class GaussianProcess(object):
                 self.mace = mace_off(model=self.mace_param.get('model'),
                                      device=self.device)
 
+        self.atoms_xyz_mask = atoms_mask.to(self.device)
+        self.Nmask = self.atoms_xyz_mask.shape[0]  # 3 * Natoms or 3 * Nreduced_atoms
+        self.atoms_mask = (self.atoms_xyz_mask[self.atoms_xyz_mask % 3 == 0] // 3).to(self.device)
+
+        # train_fp: (Ndata, Ncenter, Ndescriptor)
+        # train_dfp_dr: (Ndata, Ncenter, Natom, 3, Ndescriptor)
         self.train_fp, self.train_dfp_dr = self.generate_descriptor(self.images)
+
         self.Y = function  # Y = [Ntrain]
         self.dY = derivative  # dY = [Ntrain, Natom, 3]
-        self.model_vector = torch.empty((self.Ntrain * (1 + 3 * self.Natom),), dtype=self.torch_data_type,
+
+        self.dY = self.dY.contiguous().view(self.Ntrain, self.Natom * 3)
+        self.dY = self.dY[:, self.atoms_xyz_mask]  # shape: (Ntrain, Nselected)
+
+        # Step 3: Reshape back to (Ntrain, Nreduced_atoms, 3)
+        assert self.dY.shape[1] % 3 == 0, "Selected size must be divisible by 3"
+        Nreduced_atoms = self.atoms_mask.shape[0]
+        self.dY = self.dY.view(self.Ntrain, Nreduced_atoms, 3)
+
+        self.model_vector = torch.empty((self.Ntrain * (1 + self.Nmask),), dtype=self.torch_data_type,
                                         device=self.device)
 
         if prior is None:
-            self.prior = ConstantPrior(0.0, dtype=self.torch_data_type, device=self.device, atoms_mask=self.atoms_mask)
+            self.prior = ConstantPrior(0.0, dtype=self.torch_data_type, device=self.device,
+                                       atoms_xyz_mask=self.atoms_xyz_mask)
         else:
             self.prior = torch.tensor(prior, dtype=self.torch_data_type, device=self.device)
         self.prior_update = prior_update
@@ -325,6 +436,7 @@ class GaussianProcess(object):
         self.kernel = FPKernel(species=self.species,
                                pbc=self.pbc,
                                Natom=self.Natom,
+                               Nmask=self.Nmask,
                                kerneltype=self.kerneltype,
                                data_type=self.data_type,
                                device=self.device)
@@ -369,27 +481,50 @@ class GaussianProcess(object):
                                                return_descriptor=True,
                                                n_jobs=self.soap_param.get('n_jobs'))
 
-            dfp_dr = torch.as_tensor(dfp_dr, dtype=self.torch_data_type).to(
-                self.device)  # (Ndata, Ncenters, Natom, 3, Natom*3)
-            fp = torch.as_tensor(fp, dtype=self.torch_data_type).to(self.device)  # (Ndata, Ncenters, Natom*3)
+            fp = torch.as_tensor(fp, dtype=self.torch_data_type, device=self.device)  # (Ndata, Ncenters, Ndescriptor)
+            dfp_dr = torch.as_tensor(dfp_dr, dtype=self.torch_data_type,
+                                     device=self.device)  # (Ndata, Ncenters, Natom, 3, Ndescriptor)
+            dfp_dr = dfp_dr[:, :, self.atoms_mask, :, :]  # (Ndata, Ncenters, Nreduced_atoms, 3, Ndescriptor)
 
         elif self.descriptor == 'mace':
 
             fp = []
             dfp_dr = []
             for image in images:
-                fp__, dfp_dr__ = numerical_descriptor_gradient(image,
-                                                               self.mace,
-                                                               delta=self.mace_param.get("delta"),
-                                                               num_layers=self.mace_param.get("num_layers"),
-                                                               n_jobs=self.mace_param.get("mace_n_jobs"),
+                fp__, dfp_dr__ = mace_descriptor_gradient(image,
+                                                          self.mace,
+                                                          atoms_mask=self.atoms_mask,
+                                                          delta=self.mace_param.get("delta"),
+                                                          invariants=self.mace_param.get("invariants"),
+                                                          num_layers=self.mace_param.get("num_layers"),
+                                                          n_jobs=self.mace_param.get("mace_n_jobs"),
+                                                          dtype=self.data_type)
+                fp.append(torch.tensor(fp__, dtype=self.torch_data_type, device=self.device))
+                dfp_dr.append(torch.tensor(dfp_dr__, dtype=self.torch_data_type, device=self.device))
+
+            fp = torch.stack(fp).to(dtype=self.torch_data_type,
+                                    device=self.device)  # (Ndata, Nreduced_atoms, Ndescriptor)
+            dfp_dr = torch.stack(dfp_dr).to(dtype=self.torch_data_type,
+                                            device=self.device)  # (Ndata, Nreduced_atoms, Nreduced_atoms, 3, Ndescriptor)
+
+        elif self.descriptor == 'chebyshev':
+
+            fp = []
+            dfp_dr = []
+            for image in images:
+                fp__, dfp_dr__ = chebyshev_descriptor_gradient(image,
+                                                               self.cheb,
+                                                               atoms_mask=self.atoms_mask,
+                                                               delta=self.cheb_param.get("delta"),
+                                                               n_jobs=self.cheb_param.get("n_jobs"),
                                                                dtype=self.data_type)
                 fp.append(torch.tensor(fp__, dtype=self.torch_data_type, device=self.device))
                 dfp_dr.append(torch.tensor(dfp_dr__, dtype=self.torch_data_type, device=self.device))
 
-            fp = torch.stack(fp).to(dtype=self.torch_data_type, device=self.device)  # (Ndata, Natom, Ndescriptor)
+            fp = torch.stack(fp).to(dtype=self.torch_data_type,
+                                    device=self.device)  # (Ndata, Nreduced_atoms, Ndescriptor)
             dfp_dr = torch.stack(dfp_dr).to(dtype=self.torch_data_type,
-                                            device=self.device)  # (Ndata, Natom, Natom, 3, Ndescriptor)
+                                            device=self.device)  # (Ndata, Nreduced_atoms, Nreduced_atoms, 3, Ndescriptor)
 
         else:
             fp = []
@@ -406,6 +541,7 @@ class GaussianProcess(object):
 
             fp = fp.unsqueeze(1)  # (Ndata, 1, Natom*3)
             dfp_dr = dfp_dr.transpose(2, 3).unsqueeze(1)  # (Ndata, 1, Natom, 3, Natom*3)
+            dfp_dr = dfp_dr[:, :, self.atoms_mask, :, :]  # (Ndata, 1, Nreduced_atoms, 3, Natom*3)
 
         return fp, dfp_dr
 
@@ -418,19 +554,36 @@ class GaussianProcess(object):
                                                return_descriptor=True,
                                                n_jobs=self.soap_param.get('n_jobs'))
 
-            dfp_dr = torch.as_tensor(dfp_dr, dtype=self.torch_data_type).to(
-                self.device)  # (Ncenters, Natom, 3, Natom*3)
-            fp = torch.as_tensor(fp, dtype=self.torch_data_type).to(self.device)  # (Ncenters, Natom*3)
+            fp = torch.as_tensor(fp, dtype=self.torch_data_type, device=self.device)  # (Ncenters, Natom*3)
+            dfp_dr = torch.as_tensor(dfp_dr, dtype=self.torch_data_type,
+                                     device=self.device)  # (Ncenters, Natom, 3, Natom*3)
+            dfp_dr = dfp_dr[:, self.atoms_mask, :, :]  # (Ncenters, Nreduced_atoms, 3, Ndescriptor)
 
         elif self.descriptor == 'mace':
-            fp, dfp_dr = numerical_descriptor_gradient(image,
-                                                       self.mace,
-                                                       delta=self.mace_param.get("delta"),
-                                                       num_layers=self.mace_param.get("num_layers"),
-                                                       n_jobs=self.mace_param.get("mace_n_jobs"),
+            fp, dfp_dr = mace_descriptor_gradient(image,
+                                                  self.mace,
+                                                  atoms_mask=self.atoms_mask,
+                                                  delta=self.mace_param.get("delta"),
+                                                  invariants=self.mace_param.get("invariants"),
+                                                  num_layers=self.mace_param.get("num_layers"),
+                                                  n_jobs=self.mace_param.get("mace_n_jobs"),
+                                                  dtype=self.data_type)
+
+            fp = torch.tensor(fp, dtype=self.torch_data_type, device=self.device)  # (Nreduced_atoms, Ndescriptor)
+            dfp_dr = torch.tensor(dfp_dr, dtype=self.torch_data_type,
+                                  device=self.device)  # (Nreduced_atoms, Nreduced_atoms, 3, Ndescriptor)
+
+        elif self.descriptor == 'chebyshev':
+            fp, dfp_dr = chebyshev_descriptor_gradient(image,
+                                                       self.cheb,
+                                                       atoms_mask=self.atoms_mask,
+                                                       delta=self.cheb_param.get("delta"),
+                                                       n_jobs=self.cheb_param.get("n_jobs"),
                                                        dtype=self.data_type)
-            fp = torch.tensor(fp, dtype=self.torch_data_type, device=self.device)  # (Natom, Ndescriptor)
-            dfp_dr = torch.tensor(dfp_dr, dtype=self.torch_data_type, device=self.device)  # (Natom, Natom, 3, Ndescriptor)
+
+            fp = torch.tensor(fp, dtype=self.torch_data_type, device=self.device)  # (Nreduced_atoms, Ndescriptor)
+            dfp_dr = torch.tensor(dfp_dr, dtype=self.torch_data_type,
+                                  device=self.device)  # (Nreduced_atoms, Nreduced_atoms, 3, Ndescriptor)
 
         else:
             fp = torch.as_tensor(image.get_positions(wrap=False).reshape(-1), dtype=self.torch_data_type).to(
@@ -440,6 +593,7 @@ class GaussianProcess(object):
 
             fp = fp.unsqueeze(0)
             dfp_dr = dfp_dr.transpose(1, 2).unsqueeze(0)
+            dfp_dr = dfp_dr[:, self.atoms_mask, :, :]  # (1, Nreduced_atoms, 3, Natom*3)
 
         return fp, dfp_dr
 
@@ -453,7 +607,7 @@ class GaussianProcess(object):
         a = torch.full((self.Ntrain, 1), self.hyper_params['noise'] * self.hyper_params['noisefactor'],
                        dtype=self.torch_data_type, device=self.device)
         noise_val = self.hyper_params['noise']
-        b = noise_val.expand(self.Ntrain, 3 * self.Natom)
+        b = noise_val.expand(self.Ntrain, self.Nmask)
 
         # reg = torch.diag(torch.cat((a, b), 1).flatten() ** 2)
         diagonal_values = torch.cat((a, b), 1).flatten() ** 2
@@ -509,9 +663,11 @@ class GaussianProcess(object):
 
                 pred, kernel = self.eval_data_batch(eval_fp, eval_dfp_dr)
                 E_hat[eval_x_indexes[i][0]:eval_x_indexes[i][1]] = pred[0:data_per_batch]
-                F_hat[eval_x_indexes[i][0]:eval_x_indexes[i][1], :] = apply_force_mask(
-                    F=pred[data_per_batch:].view(data_per_batch, -1),
-                    atoms_mask=self.atoms_mask)
+                F_hat[eval_x_indexes[i][0]:eval_x_indexes[i][1], self.atoms_xyz_mask] = pred[data_per_batch:].view(
+                    data_per_batch, -1)
+                # F_hat[eval_x_indexes[i][0]:eval_x_indexes[i][1], :] = apply_force_mask(
+                #     F=pred[data_per_batch:].view(data_per_batch, -1),
+                #     atoms_xyz_mask=self.atoms_xyz_mask)
 
             return E_hat, F_hat.view((Ntest, self.Natom, 3)), None, None
 
@@ -525,9 +681,11 @@ class GaussianProcess(object):
 
                 pred, kernel = self.eval_data_batch(eval_fp, eval_dfp_dr)
                 E_hat[eval_x_indexes[i][0]:eval_x_indexes[i][1]] = pred[0:data_per_batch]
-                F_hat[eval_x_indexes[i][0]:eval_x_indexes[i][1], :] = apply_force_mask(
-                    F=pred[data_per_batch:].view(data_per_batch, -1),
-                    atoms_mask=self.atoms_mask)
+                F_hat[eval_x_indexes[i][0]:eval_x_indexes[i][1], self.atoms_xyz_mask] = pred[data_per_batch:].view(
+                    data_per_batch, -1)
+                # F_hat[eval_x_indexes[i][0]:eval_x_indexes[i][1], :] = apply_force_mask(
+                #     F=pred[data_per_batch:].view(data_per_batch, -1),
+                #     atoms_xyz_mask=self.atoms_xyz_mask)
 
                 var = self.eval_variance_batch(get_variance=get_variance,
                                                eval_fp=eval_fp,
@@ -536,9 +694,11 @@ class GaussianProcess(object):
                 std = torch.sqrt(torch.diagonal(var))
 
                 unc_e[eval_x_indexes[i][0]:eval_x_indexes[i][1]] = std[0:data_per_batch] / self.weight
-                unc_f[eval_x_indexes[i][0]:eval_x_indexes[i][1], :] = apply_force_mask(
-                    F=std[data_per_batch:].view(data_per_batch, -1),
-                    atoms_mask=self.atoms_mask)
+                unc_f[eval_x_indexes[i][0]:eval_x_indexes[i][1], self.atoms_xyz_mask] = std[data_per_batch:].view(
+                    data_per_batch, -1)
+                # unc_f[eval_x_indexes[i][0]:eval_x_indexes[i][1], :] = apply_force_mask(
+                #     F=std[data_per_batch:].view(data_per_batch, -1),
+                #     atoms_xyz_mask=self.atoms_xyz_mask)
 
             return E_hat, F_hat.view((Ntest, self.Natom, 3)), unc_e, unc_f.view((Ntest, self.Natom, 3))
 
@@ -581,12 +741,16 @@ class GaussianProcess(object):
 
         pred, kernel = self.eval_data_per_data(eval_fp_i, eval_dfp_dr_i)
         E_hat = pred[0]
-        F_hat = apply_force_mask(F=pred[1:].view(1, -1), atoms_mask=self.atoms_mask)
+        F_hat = torch.zeros((self.Natom * 3,), dtype=self.torch_data_type, device=self.device)
+        F_hat[self.atoms_xyz_mask] = pred[1:]
+        # F_hat = apply_force_mask(F=pred[1:].view(1, -1), atoms_xyz_mask=self.atoms_xyz_mask)
 
         if not get_variance:
             return E_hat, F_hat.view((self.Natom, 3)), None
 
         else:
+            unc_f = torch.zeros((self.Natom * 3,), dtype=self.torch_data_type, device=self.device)
+
             var = self.eval_variance_per_data(get_variance=True,
                                               eval_fp_i=eval_fp_i,
                                               eval_dfp_dr_i=eval_dfp_dr_i,
@@ -594,7 +758,8 @@ class GaussianProcess(object):
 
             std = torch.sqrt(torch.diagonal(var))
             unc_e = std[0] / self.weight
-            unc_f = apply_force_mask(F=std[1:].view(1, -1), atoms_mask=self.atoms_mask)
+            unc_f[self.atoms_xyz_mask] = std[1:]
+            # unc_f = apply_force_mask(F=std[1:].view(1, -1), atoms_xyz_mask=self.atoms_xyz_mask)
 
             return E_hat, F_hat.view((self.Natom, 3)), unc_e, unc_f.view((self.Natom, 3))
 
