@@ -21,7 +21,7 @@ class GaussianProcess(nn.Module):
                  use_forces=True, images=None, function=None, derivative=None,
                  sparse=None, sparse_derivative=None, autograd=False,
                  data_type='float64', device='cpu',
-                 soap_param=None, mace_param=None, descriptor='cartesian coordinates',
+                 soap_param=None, mace_param=None, cheb_param=None, descriptor='cartesian coordinates',
                  atoms_mask=None):
         super().__init__()
 
@@ -35,6 +35,7 @@ class GaussianProcess(nn.Module):
         self.device = device
         self.soap_param = soap_param
         self.mace_param = mace_param
+        self.cheb_param = cheb_param
         self.descriptor = descriptor
         self.kerneltype = kerneltype
 
@@ -93,22 +94,16 @@ class GaussianProcess(nn.Module):
                                    Natom=self.Natom,
                                    kerneltype=self.kerneltype,
                                    data_type=self.data_type,
-                                   soap_param=self.soap_param,
-                                   mace_param=self.mace_param,
-                                   descriptor=self.descriptor,
                                    device=self.device,
-                                   atoms_mask=self.atoms_mask)
+                                   )
         else:
             self.kernel = FPKernelNoforces(species=self.species,
                                            pbc=self.pbc,
                                            Natom=self.Natom,
                                            kerneltype=self.kerneltype,
                                            data_type=self.data_type,
-                                           soap_param=self.soap_param,
-                                           mace_param=self.mace_param,
-                                           descriptor=self.descriptor,
                                            device=self.device,
-                                           atoms_mask=self.atoms_mask)
+                                           )
 
         hyper_params = dict(kerneltype=self.kerneltype,
                             scale=self.scale,
